# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateNLevMVMixedResp(Template):
    'Fits multivariate mixed response models (Normal, binomial, ordered, or unordered) for two or more higher classifications (nested or crossed), allowing for missing data and a choice of uniform or Wishart priors.'

    __version__ = '1.0.0'

    tags = [ 'Model', 'N-Level', 'Normal', 'Binomial', 'Ordered multinomial', 'Unordered multinomial', 'Multivariate response']
    engines = ['eStat']

    inputs = '''
NumLevs = Integer('Number of Classifications: ')
for i in range(0, int(NumLevs)):
    selstr = 'Classification ' + str(i + 1) + ': '
    context['C' + str(i + 1)] = IDVector(selstr)
y = DataMatrix('Responses: ')

for i in range(0, len(y)):
    context['bin'+str(i+1)] = Text('Response Type for response ' + y[i] + ': ', ['Normal', 'Binomial', 'Ordered', 'Unordered'])
    if(context['bin'+str(i+1)] == 'Ordered'):
       context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
    if(context['bin'+str(i+1)] == 'Unordered'):
       context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
       for j in range(0, context['ncats'+y[i]]-1):
          context['x'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
    else:
       context['x'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)

storeresid = Boolean('Store residuals?')

lenbeta = 0
n = 0
uind = 0
for i in range(0, len(y)):
    if(context['bin'+str(i+1)] == 'Normal'):
       context['miss'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Binomial'):
       context['misb'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Ordered'):
       context['miso'+y[i]] = ParamVector(monitor=False)
       context['alpha'+y[i]] = ParamVector(customstep=True)
       context['alpha'+y[i]].ncols = -(context['ncats'+y[i]] -1)
       context['alpha'+y[i]].set_value( map(str, range(0, (context['ncats'+y[i]] -1))) )
    if(context['bin'+str(i+1)] == 'Unordered'):
       for j in range(0, context['ncats'+y[i]]-1):
          lenbeta += len(context['x'+str(i+1)+'_'+str(j)])
          context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+ str(j)] = ParamVector(monitor=False)
          for k in range(0, int(NumLevs)):
              if storeresid:
                  context['u' + str(k) + '_' + str(uind)] = ParamVector(parents=[context['C' + str(k + 1)]], as_scalar=False)
              else:
                  context['u' + str(k) + '_' + str(uind)] = ParamVector(parents=[context['C' + str(k + 1)]], as_scalar=False, monitor=False)
          uind += 1
       context['y'+ str(i+1)] = DataMatrix(value=[])
       context['origv'+str(i+1)] = List(value = [])
       n += context['ncats'+y[i]] - 1
    else:
       lenbeta += len(context['x'+str(i+1)])
       n += 1
       for k in range(0, int(NumLevs)):
           if storeresid:
               context['u' + str(k) + '_' + str(uind)] = ParamVector(parents=[context['C' + str(k + 1)]], as_scalar=False)
           else:
               context['u' + str(k) + '_' + str(uind)] = ParamVector(parents=[context['C' + str(k + 1)]], as_scalar=False, monitor=False)
       uind += 1

if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    for i in range(0, int(NumLevs)):
        context['tau_u' + str(i)] = ParamScalar()
        context['sigma_u' + str(i)] = ParamScalar(modelled = False)
else:
    omega_e = ParamVector(customstep=True)
    omega_e.ncols = ((n+1)*n)/2
    omega_e.select('Please input starting values for Level 1 variance matrix: ')
    d_e = ParamMatrix(modelled = False, customstep=True)
    d_e.size = n
    for i in range(0, int(NumLevs)):
        context['omega_u' + str(i)] = ParamMatrix(modelled = False, customstep=True)
        context['omega_u' + str(i)].size = n
        context['d_u' + str(i)] = ParamMatrix(customstep=True)
        context['d_u' + str(i)].size= n
        context['priors' + str(i)] = Text('Priors at ' + context['C' + str(i + 1)] + ' classification: ', ['Uniform', 'Wishart'])
        if context['priors' + str(i)] == 'Wishart':
            context['R' + str(i)] = List('R matrix: ')
            context['v' + str(i)] = Integer('Degrees of Freedom:')
beta = ParamVector()
beta.ncols = lenbeta
deviance = ParamScalar(modelled=False, customstep=True)
'''

    model = '''
<% lenbeta = 0 %> \\
<% numlevs = int(NumLevs) %> \\
<% n = 0 %> \\
% for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n = n + context['ncats'+y[i]]-1 %> \\
% else:
<% n = n + 1 %> \\
% endif
% endfor
model {
    for (i in 1:length(${y[0]})) {
    
    % if n == 1:
        miss0[i] ~ dnorm(mu0[i], tau)
        mu0[i] <- ${mmult2(context['x1'], 'beta', 'i', lenbeta)}
% for i in range(0, numlevs):
+ u0_${i}[${context['C' + str(i + 1)]}[i]]
% endfor
<% lenbeta += len(context['x1']) %> \\
    % else:
        dummy[i] ~ dnormal${n}a( \\
        % for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Normal'):
miss${y[i]}[i], \\
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
misb${y[i]}[i], \\
%endif
% if (context['bin'+str(i+1)] == 'Ordered'):
miso${y[i]}[i], \\
%endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
misu${context['ncats'+y[i]]}_${y[i]}_${j}[i], \\
%endfor
%endif
        % endfor
        % for i in range(0,len(context['y'])):
%if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
mu${i}_${j}[i], \\
%endfor
%else:
mu${i}[i], \\
%endif
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_e[${i}], \\
        % endfor
d_e[${(((n+1)*n)/2) - 1}])
<% uind = 0 %> \\
        % for i in range(0, len(context['y'])):
%if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
        mu${i}_${j}[i] <- ${mmult2(context['x' + str(i+1)+'_'+str(j)], 'beta', 'i', lenbeta)} \\
% for k in range(0, numlevs):
 + u${k}_${uind}[${context['C' + str(k + 1)]}[i]]\\
% endfor
<% uind += 1 %> \\
<% lenbeta += len(context['x' + str(i+1)+ '_'+str(j)]) %> 
%endfor
%else:
	  mu${i}[i] <- ${mmult2(context['x' + str(i+1)], 'beta', 'i', lenbeta)} \\
% for k in range(0, numlevs):
 + u${k}_${uind}[${context['C' + str(k + 1)]}[i]]\\
% endfor
<% uind += 1 %> \\
<% lenbeta += len(context['x' + str(i+1)]) %> 
%endif
        % endfor

        % for i in range(0,len(context['y'])):
% if (context['bin'+str(i+1)] == 'Normal'):
        miss${y[i]}[i] ~ dflat()
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
        misb${y[i]}[i] ~ dflat()
%endif
% if (context['bin'+str(i+1)] == 'Ordered'):
        miso${y[i]}[i] ~ dflat()
%endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
        misu${context['ncats'+y[i]]}_${y[i]}_${j}[i] ~ dflat()
%endfor
%endif
        % endfor

    % endif
    }
      

% for j in range(0, numlevs): 
    for (i${j + 1} in 1:length(${context['C' + str(j + 1)]})) {     
        %if len(y) == 1 :
           u${j}_0[i] ~ dnorm(0, tau_u)
    % else:
        dummy_u${j}[i${j + 1}] ~ dnormal${n}a( \\
        % for i in range(0,n):
u${j}_${i}[i${j + 1}], \\
        % endfor
        % for i in range(0,n):
0, \\
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_u${j}[${i}], \\
        % endfor
d_u${j}[${(((n+1)*n)/2) - 1}])

        % for i in range(0,n):
        u${j}_${i}[i${j + 1}] ~ dflat()
        % endfor

    % endif
    }
% endfor

    # Priors

    
    % if n == 1:
    tau ~ dgamma(0.001, 0.001)
    sigma <- 1 / sqrt(tau)
    % for i in range(0, int(NumLevs)):
        tau_u${i} ~ dgamma(0.001, 0.001)
        sigma_u${i} <- 1 / sqrt(tau_u${i})
    % endfor
    % endif

    % for i in range(0, lenbeta):
    beta_${i} ~ dflat()
    % endfor   
    % for i in range(0,(((n+1)*n)/2)):
    dumomega${i} ~ ddummy(omega_e${i})
    omega_e${i} ~ dflat()
    % endfor

}       
'''        

    preccode = '''
<%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out

    def PyToC(var):
        if isinstance(var, list):
            return "{" + ",".join(map(PyToC, var)) + "}"
        else:
            return str(var)    

    def index(n, t):
        ind = ''
        for i in range(0, n):
            ind += '[i' + str(i + 1) + ']'
        if t != '':
            ind += '[' + t + ']'
        return ind  

%>
            
<% nvar = len(context['y']) %>\\
<% numlevs = int(NumLevs) %>\\
<% n = 0 %> \\
% for i in range(0,nvar):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n += context['ncats'+y[i]]-1 %> \\
% else:
<% n += 1 %> \\
% endif
% endfor

% for var in range(0,nvar):
% if (context['bin'+str(var+1)] == 'Ordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% if (context['bin'+str(var+1)] == 'Unordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% endfor
    {
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+y[i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+y[i]]))};

            for(int j=0; j<length(${y[i]}); j++) {
                if(${y[i]}[j] <= -9.999e29) {
                } else {
                    if(miso${y[i]}[j] < low[int(${y[i]}[j])])
                        low[int(${y[i]}[j])] = miso${y[i]}[j];
                    if(miso${y[i]}[j] > upp[int(${y[i]}[j])])
                        upp[int(${y[i]}[j])] = miso${y[i]}[j];
                }
            } 
            for(int j = 1; j < ncats${y[i]}-1; j++) {
                double ul,ll;
                if(upp[j] > double(alpha${y[i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${y[i]}[j-1]);
                if(j == (ncats${y[i]} - 2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${y[i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${y[i]}[j+1]);
                alpha${y[i]}[j]= ll + (ul - ll) * random01();
            }
       }
% endif
% endfor
        // Bill's attempt of step for d_e
        SymMatrix va(${n});
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        va(${i}, ${j}) = omega_e[${count}];
<% count += 1 %>\\
% endfor
% endfor
        bool fail = false;
        mat_d_e = matrix_sym_inverse(va, fail);
        // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb(${n});
        for(int i = 0; i < length(${y[0]}); i++) {
<% numbeta = 0 %>\\
<% row = 0 %>\\
<% uind = 0 %>\\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Normal'):
            double e${row} = double(miss${y[i]}[i]) - (\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
            double e${row} = double(misb${y[i]}[i]) - (\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
            double e${row} = double(miso${y[i]}[i]) - (\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
            double e${row} = double(misu${context['ncats'+y[i]]}_${y[i]}_${j}[i]) - (\\
<% row += 1 %>\\
% for m in range(0, len(context['x' + str(i + 1) + '_' + str(j)])):
% if m != 0:
 + \\
% endif
double(${context['x' + str(i + 1) + '_' + str(j)][m]}[i]) * beta[${numbeta}] \\
<% numbeta += 1 %>\\
% endfor
)\\
% for k in range(0, numlevs):
 - double(u${k}_${uind}[int(${context['C' + str(k + 1)]}[i])]) \\
% endfor
<% uind += 1 %>\\
;
% endfor
%else :
% for m in range(0, len(context['x' + str(i+1)])):
% if m != 0:
 + \\
% endif
double(${context['x' + str(i+1)][m]}[i]) * beta[${numbeta}] \\
<% numbeta += 1 %>\\
% endfor
)\\
% for k in range(0, numlevs):
 - double(u${k}_${uind}[int(${context['C' + str(k + 1)]}[i])]) \\
% endfor
<% uind += 1 %>\\
;
%endif
% endfor
% for i in range(0, n):
% for j in range(i, n):
            sb(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }
        // Now loop over elements of va (omega_e)
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, n):
<% var2 = var1 %>\\
<% withinvar2 = withinvar1 %>\\
% for j in range(i, n):
        {
% if var1 == var2 and (context['bin'+str(var1+1)] != 'Normal'):
            omega_e_accept[${count}] = 44;
% else:
            double llcurr = length(${y[0]}) * log(matrix_determinant(mat_d_e))/2;
            for(int i1 = 0; i1 < ${n}; i1++)
                for(int j1 = 0; j1 < ${n}; j1++)
                    llcurr -= sb(i1, j1) * mat_d_e(i1, j1)/2;
            double old = va(${i}, ${j});
            va(${i}, ${j}) = dnormsd(va(${i}, ${j}), omega_e_sd[${count}]);
            mat_d_e = matrix_sym_inverse(va, fail);
            if (fail == true) {
                va(${i}, ${j}) = old;
                mat_d_e = matrix_sym_inverse(va, fail);
            } else {
                double llnew = length(${y[0]}) * log(matrix_determinant(mat_d_e))/2;
                for(int i1 = 0; i1 < ${n}; i1++)
                    for(int j1 = 0; j1 < ${n}; j1++)
                        llnew -= sb(i1, j1) * mat_d_e(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                    if(random01() > exp(llnew - llcurr))
                        change = 0;
                }
                if(change == 0) {
                    va(${i}, ${j}) = old;
                    mat_d_e = matrix_sym_inverse(va, fail);
                } else
                    omega_e_accept[${count}]++;
            }
% endif
<% count += 1 %>\\
        }
% if (context['bin'+str(var2+1)] != 'Unordered'):
<% var2+= 1 %>\\
% elif withinvar2 != (context['ncats'+y[var2]]-2):
<% withinvar2 += 1 %>\\
% else:
<% var2+= 1 %>\\
<% withinvar2 = 0 %>\\
% endif
% endfor
% if (context['bin'+str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+y[var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif
% endfor
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        omega_e[${count}] = va(${i}, ${j});
<% count += 1 %>\\
% endfor
% endfor

% for k in range(0, numlevs):
        std::vector<double*> tmp_u${k};
% for i in range(0, n):
        tmp_u${k}.push_back(u${k}_${i});
% endfor
        RectMatrix mat_u${k}(tmp_u${k}, length(u0_0));

        SymMatrix sbu${k} = mat_u${k}.T() * mat_u${k};

% if context['priors' + str(k)] == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbu${k}(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if context['priors' + str(k)] == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(context['R' + str(k)].name[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(i, n):
        sbu${k}(${i}, ${j}) += ${str(Rmat[i, j] * float(context['v' + str(k)]))};
% endfor
% endfor
% endif
        matrix_sym_invinplace(sbu${k});
% if context['priors' + str(k)] == 'Uniform':
        int vwu${k} = length(u${k}_0) - ${n + 1};
% endif
% if context['priors' + str(k)] == 'Wishart':
        int vwu${k} = length(u${k}_0) + ${context['v' + str(k)]};
% endif

        mat_d_u${k} = dwishart(vwu${k}, sbu${k}, fail);
        mat_omega_u${k} = matrix_sym_inverse(mat_d_u${k}, fail);
% endfor
    }
'''

    preparedata = '''
mydata = data['datafile']
numlevs = int(NumLevs)

for i in range(0, len(y)):
    if (context['bin'+str(i+1)] == 'Normal'):
        context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Binomial'):
        context['misb'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Ordered'):
        context['miso'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+y[i]]-1):
            context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+str(j)].ncols = -1*len(mydata.variables[y[i]]['data'])
        context['origv'+str(i+1)].name.append(y[i]) 
        for var in context['origv'+str(i+1)]:
            uniqvals = list(set(mydata.variables[var]['data'].compressed()))
            uniqvals.sort()
            uniqvals.remove(uniqvals[len(uniqvals)-1])
            for j in range(0, len(uniqvals)):
                mydata.addvariable(var + '_' + str(int(j)), data = (mydata.variables[var]['data'][:] == uniqvals[j]).astype(float))
                context['y'+str(i+1)].name.append(var + '_' + str(int(j)))
'''
